/*
 * ble_whts.c
 *
 *  Created on: Dec 16, 2015
 *      Author: C9917732
 */

#include <string.h>

#include "ble_whts.h"
#include "nordic_common.h"

#define BLE_WHTS_MAX_DATA_LEN         (GATT_MTU_SIZE_DEFAULT - 3) /**< Maximum length of data */
#define BLE_WHTS_MAX_SENSOR_CHAR_LEN  BLE_WHTS_MAX_DATA_LEN
#define BLE_WHTS_MAX_LED_CHAR_LEN     BLE_WHTS_MAX_DATA_LEN

#define INVALID_TEMP_LEVEL   0xFFFF

#define INITIAL_LED_CHARACTERISTIC_VALUE 0x0

// 021849a0-b8ac-11e5-a378-0002a5d5c51b
static const ble_uuid128_t m_base_uuid128 = WHTS_BASE_UUID;

/**@brief Function for handling the Connect event.
 *
 * @param[in]   p_whts      water heater temp svc structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_connect(ble_whts_t* p_whts, ble_evt_t* p_ble_evt)
{
    p_whts->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
}

/**@brief Function for handling the Disconnect event.
 *
 * @param[in]   p_whts      water heater temp svc structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_disconnect(ble_whts_t* p_whts, ble_evt_t* p_ble_evt)
{
    UNUSED_PARAMETER(p_ble_evt);
    p_whts->conn_handle = BLE_CONN_HANDLE_INVALID;
}

static void on_write(ble_whts_t* p_whts, ble_evt_t* p_ble_evt)
{
    ble_gatts_evt_write_t* p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;

    if ((p_evt_write->handle == p_whts->led_handles.value_handle) && (2 == p_evt_write->len))
    {
        ble_whts_evt_t evt;
        evt.evt_type       = BLE_WHTS_EVT_SET_LED_CFG;
        evt.params.led_cfg = uint16_decode(p_evt_write->data);

        p_whts->evt_handler(p_whts, &evt);
    }
}

void ble_whts_on_ble_evt(ble_whts_t* p_whts, ble_evt_t * p_ble_evt)
{
    // If these are null, wouldn't an Assert be appropriate?
    if ((NULL == p_whts) || (NULL == p_ble_evt))
    {
        return;
    }

    switch (p_ble_evt->header.evt_id)
    {
    case BLE_GAP_EVT_CONNECTED:
        on_connect(p_whts, p_ble_evt);
        break;
    case BLE_GAP_EVT_DISCONNECTED:
        on_disconnect(p_whts, p_ble_evt);
        break;
    case BLE_GATTS_EVT_WRITE:
        on_write(p_whts, p_ble_evt);
        break;
    default:
        break;
    }

}

static uint32_t whts_measurement_sensor_char_add(ble_whts_t* p_whts,
                                                 const ble_whts_init_t* p_whts_init)
{
    ble_gatts_char_md_t  char_md;
    ble_gatts_attr_md_t  cccd_md;
    ble_gatts_attr_t     attr_char_value;
    ble_uuid_t           ble_uuid;
    ble_gatts_attr_md_t  attr_md;
    uint16_t             initial_temp_value;

    if (p_whts->is_notification_supported)
    {
        memset(&cccd_md, 0, sizeof(cccd_md));

        BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
        BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);

        cccd_md.vloc = BLE_GATTS_VLOC_STACK;
    }

    memset(&char_md, 0, sizeof(char_md));

    char_md.char_props.read   = 1;
    char_md.char_props.notify = (p_whts->is_notification_supported) ? 1 : 0;
    char_md.p_char_user_desc  = NULL;
    char_md.p_char_pf         = NULL;
    char_md.p_user_desc_md    = NULL;
    char_md.p_cccd_md         = (p_whts->is_notification_supported) ? &cccd_md : NULL;
    char_md.p_sccd_md         = NULL;

    ble_uuid.type = p_whts->uuid_type;
    ble_uuid.uuid = BLE_UUID_WHTS_SENSOR;

    memset(&attr_md, 0, sizeof(attr_md));

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);

    attr_md.vloc    = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth = 0;
    attr_md.wr_auth = 0;
    attr_md.vlen    = 0;

    initial_temp_value = p_whts_init->initial_temp_level;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid    = &ble_uuid;
    attr_char_value.p_attr_md = &attr_md;
    attr_char_value.init_len  = sizeof(uint16_t); // TODO change this!
    attr_char_value.init_offs = 0;
    attr_char_value.max_len   = sizeof(uint16_t);
    attr_char_value.p_value   = (uint8_t*) &initial_temp_value;

    return sd_ble_gatts_characteristic_add(p_whts->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_whts->sensor_handles);
}

static uint32_t whts_led_ctrl_char_add(ble_whts_t* p_whts,
                                       const ble_whts_init_t* p_whts_init)
{
    ble_gatts_char_md_t  char_md;
    ble_gatts_attr_t     attr_char_value;
    ble_uuid_t           ble_uuid;
    ble_gatts_attr_md_t  attr_md;
    uint16_t             initial_value;
    uint8_t              p_encoded_val[2];

    memset(&char_md, 0, sizeof(char_md));

    char_md.char_props.write_wo_resp = 1;
    char_md.p_char_user_desc         = NULL;
    char_md.p_char_pf                = NULL;
    char_md.p_user_desc_md           = NULL;
    char_md.p_cccd_md                = NULL;
    char_md.p_sccd_md                = NULL;

    ble_uuid.type = p_whts->uuid_type;
    ble_uuid.uuid = BLE_UUID_WHTS_LED;

    memset(&attr_md, 0, sizeof(attr_md));

    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS(&attr_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_ENC_NO_MITM(&attr_md.write_perm);

    attr_md.vloc    = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth = 0;
    attr_md.wr_auth = 0;
    attr_md.vlen    = 0;

    memset(&attr_char_value, 0, sizeof(attr_char_value));
    initial_value = INITIAL_LED_CHARACTERISTIC_VALUE;

    attr_char_value.p_uuid    = &ble_uuid;
    attr_char_value.p_attr_md = &attr_md;
    attr_char_value.init_len  = sizeof(uint16_t);
    attr_char_value.init_offs = 0;
    attr_char_value.max_len   = sizeof(uint16_t);
    uint16_encode(initial_value, p_encoded_val);
    attr_char_value.p_value   = p_encoded_val;

    return sd_ble_gatts_characteristic_add(p_whts->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_whts->led_handles);
}

uint32_t ble_whts_init(ble_whts_t* p_whts, const ble_whts_init_t* p_whts_init)
{
	uint32_t err_code;
	ble_uuid_t whts_uuid;

	p_whts->evt_handler = p_whts_init->evt_handler;
	p_whts->conn_handle = BLE_CONN_HANDLE_INVALID;
	p_whts->is_notification_supported = p_whts_init->support_notification;
	p_whts->temp_value_last = INVALID_TEMP_LEVEL;

	// Add service
	//   add custom base UUID
	//
	err_code = sd_ble_uuid_vs_add(&m_base_uuid128, &p_whts->uuid_type);

	if (NRF_SUCCESS == err_code)
	{
	    whts_uuid.type = p_whts->uuid_type;
	    whts_uuid.uuid = BLE_UUID_WHTS_SERVICE;

	    err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY,
	                                        &whts_uuid,
	                                        &p_whts->service_handle);
	    if (NRF_SUCCESS == err_code)
	    {
	        // Add measurement characteristic
	        //
	        err_code = whts_measurement_sensor_char_add(p_whts, p_whts_init);

	        if (NRF_SUCCESS == err_code)
	        {
	            err_code = whts_led_ctrl_char_add(p_whts, p_whts_init);
	        }
	    }
	}

	return err_code;
}

uint32_t ble_whts_temp_sensor_update(ble_whts_t * p_whts, uint16_t temp_val)
{
    if (p_whts == NULL)
    {
        return NRF_ERROR_NULL;
    }

    uint32_t err_code = NRF_SUCCESS;
    ble_gatts_value_t gatts_value;
    uint8_t p_encoded_val[2];

    if (temp_val != p_whts->temp_value_last)
    {
        uint16_encode(temp_val, p_encoded_val);

        // Initialize value struct.
        memset(&gatts_value, 0, sizeof(gatts_value));

        gatts_value.len     = sizeof(uint16_t);
        gatts_value.offset  = 0;
        gatts_value.p_value = p_encoded_val;

        // Update database.
        err_code = sd_ble_gatts_value_set(p_whts->conn_handle,
                                          p_whts->sensor_handles.value_handle,
                                          &gatts_value);
        if (err_code == NRF_SUCCESS)
        {
            // Save new battery value.
            p_whts->temp_value_last = temp_val;
        }
        else
        {
            return err_code;
        }

        // Send value if connected and notifying.
        if ((p_whts->conn_handle != BLE_CONN_HANDLE_INVALID) && p_whts->is_notification_supported)
        {
            ble_gatts_hvx_params_t hvx_params;

            memset(&hvx_params, 0, sizeof(hvx_params));

            hvx_params.handle = p_whts->sensor_handles.value_handle;
            hvx_params.type   = BLE_GATT_HVX_NOTIFICATION;
            hvx_params.offset = gatts_value.offset;
            hvx_params.p_len  = &gatts_value.len;
            hvx_params.p_data = gatts_value.p_value;

            err_code = sd_ble_gatts_hvx(p_whts->conn_handle, &hvx_params);
        }
        else
        {
            err_code = NRF_ERROR_INVALID_STATE;
        }
    }

    return err_code;

}
